@@ -52,7 +52,7 @@
 #include "adapter.h"
 #include "device.h"
 #include "dbus-common.h"
-#include "dbus-hci.h"
+#include "event.h"
 #include "error.h"
 #include "glib-helper.h"
 #include "agent.h"
@@ -64,6 +64,9 @@
 #define DISCONNECT_TIMER	2
 #define DISCOVERY_TIMER		2
 
+/* When all services should trust a remote device */
+#define GLOBAL_TRUST "[all]"
+
 struct btd_driver_data {
 	guint id;
 	struct btd_device_driver *driver;
@@ -113,6 +116,7 @@ struct btd_device {
 	char		*alias;
 	struct btd_adapter	*adapter;
 	GSList		*uuids;
+	GSList		*services;		/* Primary services path */
 	GSList		*drivers;		/* List of driver_data */
 	GSList		*watches;		/* List of disconnect_data */
 	gboolean	temporary;
@@ -178,7 +182,8 @@ static inline DBusMessage *no_such_adapter(DBusMessage *msg)
 
 static inline DBusMessage *in_progress(DBusMessage *msg, const char *str)
 {
-	return g_dbus_create_error(msg, ERROR_INTERFACE ".InProgress", str);
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".InProgress",
+								"%s", str);
 }
 
 static void browse_request_free(struct browse_req *req)
@@ -229,6 +234,9 @@ static void device_free(gpointer user_data)
 				agent_is_busy(agent, device->authr)))
 		agent_cancel(agent);
 
+	g_slist_foreach(device->services, (GFunc) g_free, NULL);
+	g_slist_free(device->services);
+
 	g_slist_foreach(device->uuids, (GFunc) g_free, NULL);
 	g_slist_free(device->uuids);
 
@@ -270,7 +278,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	DBusMessageIter dict;
 	bdaddr_t src;
 	char name[MAX_NAME_LENGTH + 1], srcaddr[18], dstaddr[18];
-	char **uuids;
+	char **str;
 	const char *ptr;
 	dbus_bool_t boolean;
 	uint32_t class;
@@ -342,11 +350,18 @@ static DBusMessage *get_properties(DBusConnection *conn,
 				&boolean);
 
 	/* UUIDs */
-	uuids = g_new0(char *, g_slist_length(device->uuids) + 1);
+	str = g_new0(char *, g_slist_length(device->uuids) + 1);
 	for (i = 0, l = device->uuids; l; l = l->next, i++)
-		uuids[i] = l->data;
-	dict_append_array(&dict, "UUIDs", DBUS_TYPE_STRING, &uuids, i);
-	g_free(uuids);
+		str[i] = l->data;
+	dict_append_array(&dict, "UUIDs", DBUS_TYPE_STRING, &str, i);
+	g_free(str);
+
+	/* Services */
+	str = g_new0(char *, g_slist_length(device->services) + 1);
+	for (i = 0, l = device->services; l; l = l->next, i++)
+		str[i] = l->data;
+	dict_append_array(&dict, "Services", DBUS_TYPE_OBJECT_PATH, &str, i);
+	g_free(str);
 
 	/* Adapter */
 	ptr = adapter_get_path(adapter);
@@ -381,7 +396,7 @@ static DBusMessage *set_alias(DBusConnection *conn, DBusMessage *msg,
 	if (err < 0)
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".Failed",
-				strerror(-err));
+				"%s", strerror(-err));
 
 	g_free(device->alias);
 	device->alias = g_str_equal(alias, "") ? NULL : g_strdup(alias);
@@ -413,7 +428,7 @@ static DBusMessage *set_trust(DBusConnection *conn, DBusMessage *msg,
 	if (err < 0)
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".Failed",
-				strerror(-err));
+				"%s", strerror(-err));
 
 	device->trusted = value;
 
@@ -438,55 +453,30 @@ static void driver_remove(struct btd_driver_data *driver_data,
 static gboolean do_disconnect(gpointer user_data)
 {
 	struct btd_device *device = user_data;
-	disconnect_cp cp;
-	int dd;
-	uint16_t dev_id = adapter_get_dev_id(device->adapter);
 
 	device->disconn_timer = 0;
 
-	dd = hci_open_dev(dev_id);
-	if (dd < 0)
-		goto fail;
+	btd_adapter_disconnect_device(device->adapter, device->handle);
 
-	memset(&cp, 0, sizeof(cp));
-	cp.handle = htobs(device->handle);
-	cp.reason = HCI_OE_USER_ENDED_CONNECTION;
-
-	hci_send_cmd(dd, OGF_LINK_CTL, OCF_DISCONNECT,
-			DISCONNECT_CP_SIZE, &cp);
-
-	close(dd);
-
-fail:
 	return FALSE;
 }
 
 static int device_block(DBusConnection *conn, struct btd_device *device)
 {
-	int dev_id, dd, err;
+	int err;
 	bdaddr_t src;
 
 	if (device->blocked)
 		return 0;
 
-	dev_id = adapter_get_dev_id(device->adapter);
-
-	dd = hci_open_dev(dev_id);
-	if (dd < 0)
-		return -errno;
-
 	if (device->handle)
 		do_disconnect(device);
 
 	g_slist_foreach(device->drivers, (GFunc) driver_remove, device);
 
-	if (ioctl(dd, HCIBLOCKADDR, &device->bdaddr) < 0) {
-		err = -errno;
-		hci_close_dev(dd);
+	err = btd_adapter_block_address(device->adapter, &device->bdaddr);
+	if (err < 0)
 		return err;
-	}
-
-	hci_close_dev(dd);
 
 	device->blocked = TRUE;
 
@@ -507,25 +497,15 @@ static int device_block(DBusConnection *conn, struct btd_device *device)
 static int device_unblock(DBusConnection *conn, struct btd_device *device,
 							gboolean silent)
 {
-	int dev_id, dd, err;
+	int err;
 	bdaddr_t src;
 
 	if (!device->blocked)
 		return 0;
 
-	dev_id = adapter_get_dev_id(device->adapter);
-
-	dd = hci_open_dev(dev_id);
-	if (dd < 0)
-		return -errno;
-
-	if (ioctl(dd, HCIUNBLOCKADDR, &device->bdaddr) < 0) {
-		err = -errno;
-		hci_close_dev(dd);
+	err = btd_adapter_unblock_address(device->adapter, &device->bdaddr);
+	if (err < 0)
 		return err;
-	}
-
-	hci_close_dev(dd);
 
 	device->blocked = FALSE;
 
@@ -795,7 +775,6 @@ static void bonding_request_cancel(struct bonding_req *bonding)
 
 void device_request_disconnect(struct btd_device *device, DBusMessage *msg)
 {
-	GSList *l;
 	DBusConnection *conn = get_dbus_connection();
 
 	if (device->bonding)
@@ -811,21 +790,17 @@ void device_request_disconnect(struct btd_device *device, DBusMessage *msg)
 	if (device->disconn_timer)
 		return;
 
-	l = device->watches;
-	while (l) {
-		struct btd_disconnect_data *data = l->data;
-
-		l = l->next;
+	while (device->watches) {
+		struct btd_disconnect_data *data = device->watches->data;
 
 		if (data->watch)
 			/* temporary is set if device is going to be removed */
 			data->watch(device, device->temporary,
-					data->user_data);
-	}
+							data->user_data);
 
-	g_slist_foreach(device->watches, (GFunc) g_free, NULL);
-	g_slist_free(device->watches);
-	device->watches = NULL;
+		device->watches = g_slist_remove(device->watches, data);
+		g_free(data);
+	}
 
 	device->disconn_timer = g_timeout_add_seconds(DISCONNECT_TIMER,
 						do_disconnect, device);
@@ -887,7 +862,7 @@ static void device_set_connected(struct btd_device *device,
 
 		device->secmode3 = FALSE;
 
-		hcid_dbus_bonding_process_complete(&sba, &device->bdaddr, 0);
+		btd_event_bonding_process_complete(&sba, &device->bdaddr, 0);
 	}
 }
 
@@ -1056,11 +1031,10 @@ void device_get_name(struct btd_device *device, char *name, size_t len)
 	strncpy(name, device->name, len);
 }
 
-static void device_remove_bonding(struct btd_device *device)
+void device_remove_bonding(struct btd_device *device)
 {
 	char filename[PATH_MAX + 1];
 	char srcaddr[18], dstaddr[18];
-	int dd, dev_id;
 	bdaddr_t bdaddr;
 
 	adapter_get_address(device->adapter, &bdaddr);
@@ -1073,16 +1047,7 @@ static void device_remove_bonding(struct btd_device *device)
 	/* Delete the link key from storage */
 	textfile_casedel(filename, dstaddr);
 
-	dev_id = adapter_get_dev_id(device->adapter);
-
-	dd = hci_open_dev(dev_id);
-	if (dd < 0)
-		return;
-
-	/* Delete the link key from the Bluetooth chip */
-	hci_delete_stored_link_key(dd, &device->bdaddr, 0, HCI_REQ_TIMEOUT);
-
-	hci_close_dev(dd);
+	btd_adapter_remove_bonding(device->adapter, &device->bdaddr);
 }
 
 static void device_remove_stored(struct btd_device *device)
@@ -1099,6 +1064,7 @@ static void device_remove_stored(struct btd_device *device)
 	delete_entry(&src, "profiles", addr);
 	delete_entry(&src, "trusts", addr);
 	delete_all_records(&src, &device->bdaddr);
+	delete_device_service(&src, &device->bdaddr);
 
 	if (device->blocked)
 		device_unblock(conn, device, TRUE);
@@ -1340,7 +1306,7 @@ static void services_changed(struct btd_device *device)
 		uuids[i] = l->data;
 
 	emit_array_property_changed(conn, device->path, DEVICE_INTERFACE,
-					"UUIDs", DBUS_TYPE_STRING, &uuids);
+					"UUIDs", DBUS_TYPE_STRING, &uuids, i);
 
 	g_free(uuids);
 }
@@ -1812,6 +1778,8 @@ static void bonding_request_free(struct bonding_req *bonding)
 
 	device->bonding = NULL;
 
+	adapter_resume_discovery(device->adapter);
+
 	if (!device->agent)
 		return;
 
@@ -1878,57 +1846,28 @@ proceed:
 	bonding->conn = dbus_connection_ref(conn);
 	bonding->msg = dbus_message_ref(msg);
 
+	adapter_suspend_discovery(device->adapter);
+
 	return bonding;
 }
 
-static int device_authentication_requested(struct btd_device *device,
-						int handle)
+static uint8_t device_authentication_requested(struct btd_device *device,
+								int handle)
 {
-	struct hci_request rq;
-	auth_requested_cp cp;
-	evt_cmd_status rp;
-	int dd;
+	uint8_t status;
+	int err;
 
-	dd = hci_open_dev(adapter_get_dev_id(device->adapter));
-	if (dd < 0) {
-		int err = -errno;
-		error("Unable to open adapter: %s(%d)", strerror(-err), -err);
-		return err;
-	}
-
-	memset(&rp, 0, sizeof(rp));
-
-	memset(&cp, 0, sizeof(cp));
-	cp.handle = htobs(handle);
-
-	memset(&rq, 0, sizeof(rq));
-	rq.ogf    = OGF_LINK_CTL;
-	rq.ocf    = OCF_AUTH_REQUESTED;
-	rq.cparam = &cp;
-	rq.clen   = AUTH_REQUESTED_CP_SIZE;
-	rq.rparam = &rp;
-	rq.rlen   = EVT_CMD_STATUS_SIZE;
-	rq.event  = EVT_CMD_STATUS;
-
-	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
-		int err = -errno;
-		error("Unable to send HCI request: %s (%d)",
-					strerror(-err), -err);
-		hci_close_dev(dd);
-		return err;
-	}
-
-	if (rp.status) {
-		error("HCI_Authentication_Requested failed with status 0x%02x",
-				rp.status);
-		hci_close_dev(dd);
-		return rp.status;
+	err = btd_adapter_request_authentication(device->adapter, handle,
+								&status);
+	if (err < 0) {
+		error("Sending authentication request failed: %s (%d)",
+							strerror(-err), -err);
+		return HCI_UNSPECIFIED_ERROR;
 	}
 
 	info("Authentication requested");
 
-	hci_close_dev(dd);
-	return 0;
+	return status;
 }
 
 static void bonding_connect_cb(GIOChannel *io, GError *err, gpointer user_data)
@@ -2034,7 +1973,7 @@ DBusMessage *device_create_bonding(struct btd_device *device,
 		DBusMessage *reply;
 		reply = g_dbus_create_error(msg,
 				ERROR_INTERFACE ".ConnectionAttemptFailed",
-				err->message);
+				"%s", err->message);
 		error("bt_io_connect: %s", err->message);
 		g_error_free(err);
 		return reply;
@@ -2356,6 +2295,14 @@ void device_set_renewed_key(struct btd_device *device, gboolean renewed)
 	device->renewed_key = renewed;
 }
 
+void device_add_service(struct btd_device *device, const char *path)
+{
+	if (g_slist_find_custom(device->services, path, (GCompareFunc) strcmp))
+		return;
+
+	device->services = g_slist_append(device->services, g_strdup(path));
+}
+
 void btd_device_add_uuid(struct btd_device *device, const char *uuid)
 {
 	GSList *uuid_list;
